﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO.Ports;

namespace ModbusDemo
{
    public partial class ComPort : UserControl
    {
        public ComPort()
        {
            InitializeComponent();
            System.Windows.Forms.Control.CheckForIllegalCrossThreadCalls = false;//设置该属性 为false
        }
        private byte slaveID = 1;
        private SerialPort comPort = new SerialPort();

        private void ComPort_Load(object sender, EventArgs e)
        {
            cmbBaud.SelectedIndex = 7;
            cmbDataBit.SelectedIndex = 1;
            cmbParity.SelectedIndex = 0;
            cmbStopBit.SelectedIndex = 0;
            foreach (string s in SerialPort.GetPortNames())
                cmbPort.Items.Add(s);
            cmbPort.SelectedIndex = -1;
        }

        private void btOpenCOM_Click(object sender, EventArgs e)
        {
            if (cmbPort.Text == "")
            {
                MessageBox.Show("串口打开错误");
                return;
            }
            comPort.PortName = cmbPort.Text;
            comPort.BaudRate = int.Parse(cmbBaud.Text);
            comPort.DataBits = int.Parse(cmbDataBit.Text);
            if (cmbParity.Text.Substring(0, 1) == "0")
            {
                comPort.Parity = Parity.None;
            }
            else if (cmbParity.Text.Substring(0, 1) == "1")
            {
                comPort.Parity = Parity.Odd;
            }
            else if (cmbParity.Text.Substring(0, 1) == "2")
            {
                comPort.Parity = Parity.Even;
            }
            if (cmbStopBit.Text == "0")
            {
                comPort.StopBits = StopBits.None;
            }
            else if (cmbStopBit.Text == "1")
            {
                comPort.StopBits = StopBits.One;
            }
            try
            {
                comPort.Open();
               //comPort.Write("01 03 00 00 00 05 85 C9");//写数据
                comPort.DataReceived += new SerialDataReceivedEventHandler(post_DataReceived);//串口接收处理函数
                btOpenCOM.Enabled = false;
                btCloseCOM.Enabled = true;
                Console.WriteLine(DateTime.Now.ToString() + " =>Open " + comPort.PortName + " sucessfully!");
                //MessageBox.Show(DateTime.Now.ToString() + System.Environment.NewLine + " =>Open " + comPort.PortName + " sucessfully!");
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error: " + ex.Message);
                MessageBox.Show("Error: " + ex.Message);
                return;
            }
        }
        //字节数组转16进制字符串
        public static string byteToHexStr(byte[] bytes)
        {
            string returnStr = "";
            if (bytes != null)
            {
                for (int i = 0; i < bytes.Length; i++)
                {
                    returnStr += bytes[i].ToString("X2");
                }
            }
            return returnStr;
        }
        static int buffersize = 18;   //十六进制数的大小（假设为9Byte,可调整数字大小）
        byte[] buffer = new Byte[buffersize];   //创建缓冲区

        private void post_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            int size = 0;
            
            comPort.Read(buffer, 0, buffersize);
            string ss;
            ss = byteToHexStr(buffer); //用到函数byteToHexStr

            Console.WriteLine(comPort.ReadExisting());
            Console.WriteLine(ss);

            size = comPort.ReadBufferSize;

            if (buffer[0]==255)
            {
                modbus_Master.TepArray[0] = buffer[0];
                modbus_Master.Sign = 2;
                Console.WriteLine(comPort.ReadExisting());
            }
            else 
            {
                modbus_Master.Sign =1;
                modbus_Master.TepArray[0] = 0;
            }
        
        }

        private void btCloseCOM_Click(object sender, EventArgs e)
        {
                //Close comport first,then stop and dispose slave.
                comPort.Close();
                btOpenCOM.Enabled = true;
                btCloseCOM.Enabled = false;
                Console.WriteLine(DateTime.Now.ToString() + " =>Disconnect " + comPort.PortName);
                MessageBox.Show(DateTime.Now.ToString() + System.Environment.NewLine + " =>Disconnect " + comPort.PortName);
        }
    }
}
