﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO.Ports;
using Modbus;
using Modbus.Device;
using Modbus.Data;
using System.Threading;

namespace ModbusDemo
{
    public partial class Modbus_slave : UserControl
    {
        public Modbus_slave()
        {
            InitializeComponent();
        }
        ModbusSlave slave;
        private byte slaveID = 1;
        private SerialPort comPort = new SerialPort();
        private void Modbus_slave_Load(object sender, EventArgs e)
        {
            comboBox1.SelectedIndex = 0;
            cmbBaud.SelectedIndex = 7;
            cmbDataBit.SelectedIndex = 1;
            cmbParity.SelectedIndex = 0;
            cmbStopBit.SelectedIndex = 0;
            foreach (string s in SerialPort.GetPortNames())
                cmbPort.Items.Add(s);
            cmbPort.SelectedIndex = -1;
        }
        private void btOpenCOM_Click(object sender, EventArgs e)
        {
            comPort.PortName = cmbPort.Text;
            comPort.BaudRate = int.Parse(cmbBaud.Text);
            comPort.DataBits = int.Parse(cmbDataBit.Text);
            if (cmbParity.Text.Substring(0, 1) == "0")
            {
                comPort.Parity = Parity.None;
            }
            else if (cmbParity.Text.Substring(0, 1) == "1")
            {
                comPort.Parity = Parity.Odd;
            }
            else if (cmbParity.Text.Substring(0, 1) == "2")
            {
                comPort.Parity = Parity.Even;
            }
            if (cmbStopBit.Text == "0")
            {
                comPort.StopBits = StopBits.None;
            }
            else if (cmbStopBit.Text == "1")
            {
                comPort.StopBits = StopBits.One;
            }
            try
            {
                comPort.Open();
                slave = ModbusSerialSlave.CreateRtu(slaveID, comPort);
                slave.ModbusSlaveRequestReceived += new EventHandler<ModbusSlaveRequestEventArgs>(Modbus_Request_Event);
                slave.DataStore = Modbus.Data.DataStoreFactory.CreateDefaultDataStore();
                slave.DataStore.DataStoreWrittenTo += new EventHandler<DataStoreEventArgs>(Modbus_DataStoreWriteTo);
                //slave.DataStore.DataStoreReadFrom += new EventHandler<DataStoreEventArgs>(DataStore_DataStoreReadFrom);
                btOpenCOM.Enabled = false;
                btCloseCOM.Enabled = true;
                slave.Listen();
                timer.Enabled = true;
                Console.WriteLine(DateTime.Now.ToString() + " =>Open " + comPort.PortName + " sucessfully!");
                MessageBox.Show(DateTime.Now.ToString() + System.Environment.NewLine + " =>Open " + comPort.PortName + " sucessfully!");
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error: " + ex.Message);
                MessageBox.Show("Error: " + ex.Message);
                return;
            }
        }

        //private void DataStore_DataStoreReadFrom(object sender, DataStoreEventArgs e)
        //{
            
        //}

        private void btCloseCOM_Click(object sender, EventArgs e)
        {
            //Close comport first,then stop and dispose slave.
            comPort.Close();
            slave.Stop();
            slave.Dispose();
            btOpenCOM.Enabled = true;
            btCloseCOM.Enabled = false;
            timer.Enabled = false;
            Console.WriteLine(DateTime.Now.ToString() + " =>Disconnect " + comPort.PortName);
            MessageBox.Show(DateTime.Now.ToString() + System.Environment.NewLine + " =>Disconnect " + comPort.PortName);
        }

        
        private void Modbus_Request_Event(object sender, Modbus.Device.ModbusSlaveRequestEventArgs e)
        {
            try
            {
                CheckForIllegalCrossThreadCalls = false;
                //request from master
                byte fc = e.Message.FunctionCode;
                byte[] data = e.Message.MessageFrame;
                byte[] byteStartAddress = new byte[] { data[3], data[2] };
                byte[] byteNum = new byte[] { data[5], data[4] };
                Int16 StartAddress = BitConverter.ToInt16(byteStartAddress, 0);
                Int16 NumOfPoint = BitConverter.ToInt16(byteNum, 0);
                Console.WriteLine(fc.ToString() + "," + StartAddress.ToString() + "," + NumOfPoint.ToString());
                modbus_Master.Values = fc.ToString() + "," + StartAddress.ToString() + "," + NumOfPoint.ToString();

            }
            catch (Exception ex)
            {
                Console.WriteLine("Error: " + ex.Message);
                MessageBox.Show("Error: " + ex.Message);
                return;
            }   
        }

        private void Modbus_DataStoreWriteTo(object sender, Modbus.Data.DataStoreEventArgs e)
        {
            //this.Text = "DataType=" + e.ModbusDataType.ToString() + "  StartAdress=" + e.StartAddress;
            int iAddress = e.StartAddress;//e.StartAddress;
            switch (e.ModbusDataType)
            {
                case ModbusDataType.HoldingRegister:
                    for (int i = 0; i < e.Data.B.Count; i++)
                    {
                        //Set AO                 

                        //e.Data.B[i] already write to slave.DataStore.HoldingRegisters[e.StartAddress + i + 1]
                        //e.StartAddress starts from 0
                        //You can set AO value to hardware here

                        DoAOUpdate(iAddress, e.Data.B[i].ToString());
                        iAddress++;
                    }
                    break;

                case ModbusDataType.Coil:
                    for (int i = 0; i < e.Data.A.Count; i++)
                    {
                        //Set DO
                        //e.Data.A[i] already write to slave.DataStore.CoilDiscretes[e.StartAddress + i + 1]
                        //e.StartAddress starts from 0
                        //You can set DO value to hardware here

                        DoDOUpdate(iAddress, e.Data.A[i]);
                        iAddress++;
                        if (e.Data.A.Count == 1)
                        {
                            break;
                        }
                    }
                    break;
            }

        }
        #region "Set AO"
        private delegate void UpdateAOStatusDelegate(int index, String message);
        private void DoAOUpdate(int index, String message)
        {
            if (this.InvokeRequired)
            {
                // we were called on a worker thread
                // marshal the call to the user interface thread
                this.Invoke(new UpdateAOStatusDelegate(DoAOUpdate),
                            new object[] { index, message });
                return;
            }

            // this code can only be reached
            // by the user interface thread
            switch (index)
            {
                case 0:
                    this.txtAO1.Text = message;
                    break;
                case 1:
                    this.txtAO2.Text = message;
                    break;
                case 2:
                    this.txtAO3.Text = message;
                    break;
                case 3:
                    this.txtAO4.Text = message;
                    break;
            }

        }
        #endregion

        #region "Set DO"
        private delegate void UpdateDOStatusDelegate(int index, bool value);
        private void DoDOUpdate(int index, bool value)
        {
            if (this.InvokeRequired)
            {
                // we were called on a worker thread
                // marshal the call to the user interface thread
                this.Invoke(new UpdateDOStatusDelegate(DoDOUpdate),
                            new object[] { index, value });
                return;
            }

            // this code can only be reached
            // by the user interface thread
            switch (index)
            {
                case 0:
                    this.chkDO1.Checked = value;
                    break;
                case 1:
                    this.chkDO2.Checked = value;
                    break;
                case 2:
                    this.chkDO3.Checked = value;
                    break;
                case 3:
                    this.chkDO4.Checked = value;
                    break;
            }

        }
        #endregion

        private void timer_Tick(object sender, EventArgs e)
        {
            //update input values to datastore
            //DI
            slave.DataStore.InputDiscretes[1] = chkDI1.Checked;
            slave.DataStore.InputDiscretes[2] = chkDI2.Checked;
            slave.DataStore.InputDiscretes[3] = chkDI3.Checked;
            slave.DataStore.InputDiscretes[4] = chkDI4.Checked;
            //AI
            slave.DataStore.InputRegisters[1] = Convert.ToUInt16(txtAI1.Text);
            slave.DataStore.InputRegisters[2] = Convert.ToUInt16(txtAI2.Text);
            slave.DataStore.InputRegisters[3] = Convert.ToUInt16(txtAI3.Text);
            slave.DataStore.InputRegisters[4] = Convert.ToUInt16(txtAI4.Text);
            
            //AO
            switch (comboBox1.SelectedIndex)
            {
                case 0:
                {
                        slave.DataStore.HoldingRegisters[1] = Convert.ToUInt16(txtAO1.Text);
                        slave.DataStore.HoldingRegisters[2] = Convert.ToUInt16(txtAO2.Text);
                        slave.DataStore.HoldingRegisters[3] = Convert.ToUInt16(txtAO3.Text);
                        slave.DataStore.HoldingRegisters[4] = Convert.ToUInt16(txtAO4.Text);
                        slave.DataStore.HoldingRegisters[5] = 319; slave.DataStore.HoldingRegisters[6] = 319;
                        slave.DataStore.HoldingRegisters[7] = 319; slave.DataStore.HoldingRegisters[8] = 3;
                        slave.DataStore.HoldingRegisters[9] = 0; slave.DataStore.HoldingRegisters[10] = 0;
                        slave.DataStore.HoldingRegisters[11] = 5000; slave.DataStore.HoldingRegisters[12] = 6000;
                        slave.DataStore.HoldingRegisters[13] = 7000; slave.DataStore.HoldingRegisters[14] = 8000;
                        slave.DataStore.HoldingRegisters[15] = 9000; slave.DataStore.HoldingRegisters[16] = 0;
                        slave.DataStore.HoldingRegisters[17] = 0; slave.DataStore.HoldingRegisters[18] = 0;
                        slave.DataStore.HoldingRegisters[19] = 0; slave.DataStore.HoldingRegisters[20] = 0;
                        slave.DataStore.HoldingRegisters[21] = 0; slave.DataStore.HoldingRegisters[22] = 0;
                        slave.DataStore.HoldingRegisters[23] = 0; slave.DataStore.HoldingRegisters[24] = 0;
                        slave.DataStore.HoldingRegisters[25] = 0; slave.DataStore.HoldingRegisters[26] = 0;
                        slave.DataStore.HoldingRegisters[27] = 0; slave.DataStore.HoldingRegisters[28] = 0;
                        slave.DataStore.HoldingRegisters[29] = 0; slave.DataStore.HoldingRegisters[30] = 0;
                        slave.DataStore.HoldingRegisters[31] = 0; slave.DataStore.HoldingRegisters[32] = 0;
                        slave.DataStore.HoldingRegisters[33] = 220; slave.DataStore.HoldingRegisters[34] = 500;
                        slave.DataStore.HoldingRegisters[35] = 50; slave.DataStore.HoldingRegisters[36] = 1100;
                        slave.DataStore.HoldingRegisters[37] = 220; slave.DataStore.HoldingRegisters[38] = 500;
                        slave.DataStore.HoldingRegisters[39] = 50; slave.DataStore.HoldingRegisters[40] = 1100;
                        slave.DataStore.HoldingRegisters[41] = 5000; slave.DataStore.HoldingRegisters[42] = 5000;
                        slave.DataStore.HoldingRegisters[43] = 0; slave.DataStore.HoldingRegisters[44] = 0;
                    }
               break;
                default:

                    break;

            }
            //DO
            slave.DataStore.CoilDiscretes[1] = chkDO1.Checked;
            slave.DataStore.CoilDiscretes[2] = chkDO2.Checked;
            slave.DataStore.CoilDiscretes[3] = chkDO3.Checked;
            slave.DataStore.CoilDiscretes[4] = chkDO4.Checked;
        }
    }

}
