﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
//step1. reference nmodbuspc.dll, and using the namespaces.
using Modbus.Device;    //for modbus master
using System.IO.Ports;  //for serial port
using ModbusRTU_Master;
using ModbusDemo.windows;
using System.IO;
using System.Threading;
using System.Runtime.InteropServices;
using System.Diagnostics;
/*
   //Modbus exception codes definition                            
   * Code   * Name                                      * Meaning
     01       ILLEGAL FUNCTION                            The function code received in the query is not an allowable action for the server.

     02       ILLEGAL DATA ADDRESS                        The data addrdss received in the query is not an allowable address for the server.

     03       ILLEGAL DATA VALUE                          A value contained in the query data field is not an allowable value for the server.

     04       SLAVE DEVICE FAILURE                        An unrecoverable error occurred while the server attempting to perform the requested action.

     05       ACKNOWLEDGE                                 This response is returned to prevent a timeout error from occurring in the client (or master)
                                                          when the server (or slave) needs a long duration of time to process accepted request.

     06       SLAVE DEVICE BUSY                           The server (or slave) is engaged in processing a long–duration program command , and the
                                                          client (or master) should retransmit the message later when the server (or slave) is free.

     08       MEMORY PARITY ERROR                         The server (or slave) attempted to read record file, but detected a parity error in the memory.

     0A       GATEWAY PATH UNAVAILABLE                    The gateway is misconfigured or overloaded.

     0B       GATEWAY TARGET DEVICE FAILED TO RESPOND     No response was obtained from the target device. Usually means that the device is not present on the network.
 */
namespace ModbusDemo
{
    public partial class modbus_Master : UserControl
    {
        SerialPort serialPort = new SerialPort(); //create a new SerialPort object with default settings.
        ModbusSerialMaster master;
        List<TextBox> listAI = new List<TextBox>();
        List<TextBox> listAO = new List<TextBox>();
        List<PictureBox> listDI = new List<PictureBox>();
        List<PictureBox> listDO = new List<PictureBox>();
        string ImagefilePath =null;
        FileStream fsl;
        byte[] MNames = new byte[12];
        byte[] Ccids = new byte[20];
        byte[] state = new byte[2];
        Int32 SN;
        private int waitTime;
        private Thread t_UpgradeConn;
        private bool bStart = false;
        AutoResetEvent exitEvent;
        public delegate void invokeDelegate();
        [DllImport("user32.dll")]
        public static extern int MessageBeep(uint uType);
        public static string Values { get; set; }
        public static int Sign { get; set; }
        public static byte[] TepArray { get; set; } = new Byte[100];
        public modbus_Master()
        {
            InitializeComponent();
            Sign = 0;
        }

        private void modbus_Master_Load(object sender, EventArgs e)
        {
            cmbBaud.SelectedIndex = 7;
            cmbDataBit.SelectedIndex = 1;
            cmbParity.SelectedIndex = 0;
            cmbStopBit.SelectedIndex = 0;
            foreach (string s in SerialPort.GetPortNames())
                cmbPort.Items.Add(s);
            cmbPort.SelectedIndex = -1;

            listDI.Add(DI0);
            listDI.Add(DI1);
            listDI.Add(DI2);
            listDI.Add(DI3);
            listDI.Add(DI4);
            listDI.Add(DI5);
            listDI.Add(DI6);
            listDI.Add(DI7);
            listDI.Add(DI8);
            listDI.Add(DI9);
            listDI.Add(DI10);
            listDI.Add(DI11);
            listDI.Add(DI12);
            listDI.Add(DI13);
            listDI.Add(DI14);
            listDI.Add(DI15);
            //listDO.Add(DO0);
            //listDO.Add(DO1);
            //listDO.Add(DO2);
            //listDO.Add(DO3);

            //listAI.Add(AI0);
            //listAI.Add(AI1);
            //listAI.Add(AI2);
            //listAI.Add(AI3);

            //listAO.Add(AO0);
            //listAO.Add(AO1);
            //listAO.Add(AO2);
            //listAO.Add(AO3);
            exitEvent = new AutoResetEvent(false);
            waitTime = 6000;
            t_UpgradeConn = new Thread(UpgradeProc);
            t_UpgradeConn.IsBackground = true;
            t_UpgradeConn.Start();
        }

        private void btOpenCOM_Click(object sender, EventArgs e)
        {
            if (cmbPort.Text == "")
            {
                MessageBox.Show("串口打开错误");
                return;
            }
            else
            {
                serialPort.PortName = cmbPort.Text;
                serialPort.BaudRate = int.Parse(cmbBaud.Text);
                serialPort.DataBits = int.Parse(cmbDataBit.Text);
            }
            if (cmbParity.Text.Substring(0, 1) == "0")
            {
                serialPort.Parity = Parity.None;
            }
            else if (cmbParity.Text.Substring(0, 1) == "1")
            {
                serialPort.Parity = Parity.Odd;
            }
            else if (cmbParity.Text.Substring(0, 1) == "2")
            {
                serialPort.Parity = Parity.Even;
            }
            if (cmbStopBit.Text == "0")
            {
                serialPort.StopBits = StopBits.None;
            }
            else if (cmbStopBit.Text == "1")
            {
                serialPort.StopBits = StopBits.One;
            }
            try
            {
                serialPort.Open();
                master = ModbusSerialMaster.CreateRtu(serialPort);
                master.Transport.Retries = 0;   //don't have to do retries
                master.Transport.ReadTimeout = 6000; //milliseconds
                master.Transport.WriteTimeout = 6000;
                uint beepI = 0x00000066;
                MessageBeep(beepI);
                Console.WriteLine(DateTime.Now.ToString() + " =>Open " + serialPort.PortName + " sucessfully!");
                //MessageBox.Show(DateTime.Now.ToString()   + System.Environment.NewLine + " =>Open " + serialPort.PortName + " sucessfully!");
                btOpenCOM.Enabled = false;
                btCloseCOM.Enabled = true;
                button1.Enabled = true;
                
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error: " + ex.Message);
                MessageBox.Show("Error: " + ex.Message);
                return;
            }
        }

        private void btCloseCOM_Click(object sender, EventArgs e)
        {
            //Close comport first,then stop and dispose slave.
            serialPort.Close();
            btOpenCOM.Enabled = true;
            btCloseCOM.Enabled = false;
            timer1.Enabled = false;
            button1.Enabled = false;
            StartTest.Enabled = false;
            reset.Enabled = false;
            Console.WriteLine(DateTime.Now.ToString() + " =>Disconnect " + serialPort.PortName);
            MessageBox.Show(DateTime.Now.ToString() + System.Environment.NewLine  + " =>Disconnect " + serialPort.PortName);
        }

        //set DO
        private void DO_click(object sender, EventArgs e)
        {
            byte slaveID = 1;
            if (serialPort.IsOpen == true)
            {
                ushort index = ushort.Parse(((PictureBox)sender).Tag.ToString());
                if (((PictureBox)sender).BackColor == Color.DarkRed)
                    master.WriteSingleCoil(slaveID, index, true);
                else
                    master.WriteSingleCoil(slaveID, index, false);
            }
        }
        //set AO
        private void AO_click(object sender, EventArgs e)
        {
            byte slaveID = 1;
            frmInputValue inputvalue = new frmInputValue();
            if (serialPort.IsOpen == true)
            {
                ushort index = ushort.Parse(((TextBox)sender).Tag.ToString());
                inputvalue.StringValue = ((TextBox)sender).Text;
                inputvalue.ShowDialog();
                if (inputvalue.DialogResult == DialogResult.OK)
                {
                    double value = inputvalue.Value;
                    ushort aovalue = (ushort)value;

                    //use gain=20/32767, offset=0
                    //ushort aovalue = (ushort)(value * 32767 / 20.0);
                    master.WriteSingleRegister(slaveID, index, aovalue);
                }
            }
        }
        /*************************************************************************************/
        private void button1_Click(object sender, EventArgs e)//设备信息读取
        {
            ////下载部分
            Process proc = null;
            try
            {
                //"C:\Users\shaxu\OneDrive\C# lib\c#-utilityLibrary\ModbusDemo\bin\Debug\"
                //当前运行目录
                string targetDir = string.Format(@System.Windows.Forms.Application.StartupPath);//this is where testChange.bat lies
                proc = new Process();
                proc.StartInfo.WorkingDirectory = targetDir;
                proc.StartInfo.FileName = "jlink_app_download.bat";
                proc.StartInfo.Arguments = string.Format("10");//this is argument
                //proc.StartInfo.CreateNoWindow = true;
                //proc.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;//这里设置DOS窗口不显示，经实践可行
                proc.Start();
                proc.WaitForExit();
            }
            catch (Exception ex)
            {
                Console.WriteLine("Exception Occurred :{0},{1}", ex.Message, ex.StackTrace.ToString());
            }

            try
            {
                ushort[] register = master.ReadHoldingRegisters(1, 0, 3);
                SN = (register[0] << 16) + (Int32)register[1];
                DTUSN.Text = SN.ToString();
                

                byte[] bs = BitConverter.GetBytes(register[2]);
                version.Text = (
                                Convert.ToInt32(bs[1].ToString("X2")) / 10 + Convert.ToInt32(bs[1].ToString("X2")) % 10 + "." +
                                Convert.ToInt32(bs[0].ToString("X2")) / 10 + "." + Convert.ToInt32(bs[0].ToString("X2")) % 10
                                ).ToString();

                ushort[] Namestr = master.ReadHoldingRegisters(1, 3, 6);
                for (int i = 0; i < 4; i++)
                {
                    if (i < 2)
                        MNames[i] = (byte)((Namestr[0] >> (i * 8)) & 0x00ff);
                    else
                        MNames[i] = (byte)(Namestr[1] >> ((i % 2) * 8) & 0x00ff);
                }
                for (int i = 4; i < 8; i++)
                {
                    if (i < 6)
                        MNames[i] = (byte)((Namestr[2] >> (i * 8)) & 0x00ff);
                    else
                        MNames[i] = (byte)(Namestr[3] >> ((i % 2) * 8) & 0x00ff);
                }
                for (int i = 8; i < 12; i++)
                {
                    if (i < 10)
                        MNames[i] = (byte)((Namestr[4] >> (i * 8)) & 0x00ff);
                    else
                        MNames[i] = (byte)(Namestr[5] >> ((i % 2) * 8) & 0x00ff);
                }
                Nametype.Text = System.Text.Encoding.ASCII.GetString(MNames);

                Int32 year, month, day;
                ushort[] RtcDates = master.ReadHoldingRegisters(1, 23, 6);
                year = RtcDates[0];
                month = RtcDates[1];
                day = RtcDates[2];

                RtcDate.Text = year.ToString() + "." +
                    month.ToString() + "." +
                      day.ToString();
                if (RtcDate.Text != "1999.12.31")
                {
                    RTCLED.BackColor = Color.Lime;
                }
                else
                    RTCLED.BackColor = Color.Red;

                StartTest.Enabled = true;
                reset.Enabled = true;
                this.UpdateBtn.Enabled = true;
                this.choosebth.Enabled = true;
                this.button2.Enabled = true;
                //关闭写入功能
                DTUSN.Enabled = true;
                Nametype.Enabled = true;

            }

            catch (Exception exception)
            {
                //Connection exception
                //No response from server.
                //The server maybe close the com port, or response timeout.
                if (exception.Source.Equals("System"))
                {
                    Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                    return ;
                    //MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                }
                //The server return error code.
                //You can get the function code and exception code.
                if (exception.Source.Equals("nModbusPC"))
                {
                    string str = exception.Message;
                    int FunctionCode;
                    string ExceptionCode;

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);

                    FunctionCode = Convert.ToInt16(str.Remove(str.IndexOf("\r\n")));
                    Console.WriteLine("Function Code: " + FunctionCode.ToString("X"));
                    MessageBox.Show("Function Code: " + FunctionCode.ToString("X"));

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);
                    ExceptionCode = str.Remove(str.IndexOf("-"));
                    switch (ExceptionCode.Trim())
                    {
                        case "1":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            break;
                        case "2":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            break;
                        case "3":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            break;
                        case "4":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            break;
                    }
                }
            }
        }


        private void timer1_Tick(object sender, EventArgs e)
        {

         try
          {
            ushort[] signflag = master.ReadHoldingRegisters(1, 10, 1);

                if (Sign == 1)
                {
                    SlaveState.BackColor = Color.Lime;
                    Sign = 0;
                }
                if(Sign ==2)
                {
                    SlaveState.BackColor = Color.Red;
                    MessageBox.Show("从机通道异常，异常数据头为:" + " ‘ " + TepArray[0].ToString("X2") + " ’ ");
                    TepArray[0] = 0;
                    Sign = 0;
                }

                if (signflag[0] == 3)
                {
                /*获取状态*/
                ushort[] register = master.ReadHoldingRegisters(1, 11, 1);
                    for (int i = 0; i < 2; i++)
                {
                    state[i] = (byte)((register[0] >> (i * 8)) & 0x00ff);
                }
                bool[] status = GetBoolArr(state[0]);
                bool[] status2 = GetBoolArr(state[1]);

                for (int num = 7; num >= 0; num--)
                {
                    if (status[num] == true)
                        listDI[num].BackColor = Color.Lime;
                    else
                        listDI[num].BackColor = Color.Red;

                    if (status2[num] == true)

                        listDI[num + 8].BackColor = Color.Lime;
                    else
                        listDI[num + 8].BackColor = Color.Red;
                }
               /*CCID获取*/
               ushort[] CCID = master.ReadHoldingRegisters(1, 13, 10);
                    for (int i = 0; i < 4; i++)
                    {
                        if (i < 2)
                            Ccids[i] = (byte)((CCID[0] >> (i * 8)) & 0x00ff);
                        else
                            Ccids[i] = (byte)(CCID[1] >> ((i % 2) * 8) & 0x00ff);
                    }
                    for (int i = 4; i < 8; i++)
                    {
                        if (i < 6)
                            Ccids[i] = (byte)((CCID[2] >> (i * 8)) & 0x00ff);
                        else
                            Ccids[i] = (byte)(CCID[3] >> ((i % 2) * 8) & 0x00ff);
                    }
                    for (int i = 8; i < 12; i++)
                    {
                        if (i < 10)
                            Ccids[i] = (byte)((CCID[4] >> (i * 8)) & 0x00ff);
                        else
                            Ccids[i] = (byte)(CCID[5] >> ((i % 2) * 8) & 0x00ff);
                    }
                    for (int i = 12; i < 16; i++)
                    {
                        if (i < 14)
                            Ccids[i] = (byte)((CCID[6] >> (i * 8)) & 0x00ff);
                        else
                            Ccids[i] = (byte)(CCID[7] >> ((i % 2) * 8) & 0x00ff);
                    }
                    for (int i = 16; i < 20; i++)
                    {
                        if (i < 18)
                            Ccids[i] = (byte)((CCID[8] >> (i * 8)) & 0x00ff);
                        else
                            Ccids[i] = (byte)(CCID[9] >> ((i % 2) * 8) & 0x00ff);
                    }
                    CcidLab.Text = System.Text.Encoding.ASCII.GetString(Ccids);
                    timer1.Enabled = false;
                    DialogResult dr = MessageBox.Show("从机测试完成请确认", "提示", MessageBoxButtons.OKCancel, MessageBoxIcon.Question);
                    if (dr == DialogResult.OK)
                    {
                        //点确定的代码
                        StartTest.Enabled = true;
                        for (int num = 7; num >= 0; num--)
                        {
                            if (status[num] == true)
                                listDI[num].BackColor = Color.Navy;
                            else
                                listDI[num].BackColor = Color.Red;

                            if (status2[num] == true)

                                listDI[num + 8].BackColor = Color.Navy;
                            else
                                listDI[num + 8].BackColor = Color.Red;
                        }

                        if (Values != null)
                        {
                            DialogResult dv = MessageBox.Show("主机回复正常请确认" + "\n\r" + "回复内容：" + Values.ToString(), "提示", MessageBoxButtons.OKCancel, MessageBoxIcon.Question);
                        }
                        RTCLED.BackColor = Color.Navy;
                        SlaveState.BackColor = Color.Navy;
                        DTUSN.Text = "......";
                        SNcheck.Text = "......";
                        version.Text = "......";
                        Nametype.Text = "......";
                        CcidLab.Text = "......";
                        RtcDate.Text = ".......";
                        btOpenCOM.Enabled = false;
                        button2.Enabled = false;
                        btCloseCOM.Enabled = true;
                        timer1.Enabled = false;
                        button1.Enabled = true;
                        StartTest.Enabled = false;
                        reset.Enabled = false;
                    }
                    //else
                    //{
                        RTCLED.BackColor = Color.Navy;
                        SlaveState.BackColor = Color.Navy;
                        btOpenCOM.Enabled = false;
                        button2.Enabled = false;
                        btCloseCOM.Enabled = true;
                        timer1.Enabled = false;
                        button1.Enabled = true;
                        StartTest.Enabled = false;
                        reset.Enabled = false;
                   // }
                }
            }
            catch (Exception exception)
            {
                //Connection exception
                //No response from server.
                //The server maybe close the com port, or response timeout.
                if (exception.Source.Equals("System"))
                {
                    timer1.Enabled = false;
                    StartTest.Enabled = false;
                    reset.Enabled = false;
                    this.UpdateBtn.Enabled = false;
                    this.choosebth.Enabled = false;
                    this.button2.Enabled = false;
                    Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                    MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                    return;
                }
                //The server return error code.
                //You can get the function code and exception code.
                if (exception.Source.Equals("nModbusPC"))
                {
                    string str = exception.Message;
                    int FunctionCode;
                    string ExceptionCode;

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);
                    FunctionCode = Convert.ToInt16(str.Remove(str.IndexOf("\r\n")));
                    Console.WriteLine("Function Code: " + FunctionCode.ToString("X"));
                    MessageBox.Show("Function Code: " + FunctionCode.ToString("X"));

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);
                    ExceptionCode = str.Remove(str.IndexOf("-"));
                    switch (ExceptionCode.Trim())
                    {
                        case "1":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            break;
                        case "2":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            break;
                        case "3":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            break;
                        case "4":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            break;
                    }
                }
            }
        }
        private void button2_Click(object sender, EventArgs e)
        {
           //MessageBox.Show("开始写入测试");
            try
            {
                master.WriteSingleRegister(1, 10, 1);
                timer1.Enabled = true;
                button1.Enabled = false;
                StartTest.Enabled = false;
                reset.Enabled = false;
                button2.Enabled = false;
                this.UpdateBtn.Enabled = false;
                this.choosebth.Enabled = false;
            }
            catch (Exception exception)
            {
                //Connection exception
                //No response from server.
                //The server maybe close the com port, or response timeout.
                if (exception.Source.Equals("System"))
                {
                    Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                    MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                    StartTest.Enabled = false;
                    reset.Enabled = false;
                    this.UpdateBtn.Enabled = false;
                    this.choosebth.Enabled = false;
                    timer1.Enabled = false;
                    return;
                }
                //The server return error code.
                //You can get the function code and exception code.
                if (exception.Source.Equals("nModbusPC"))
                {
                    string str = exception.Message;
                    int FunctionCode;
                    string ExceptionCode;

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);
                    FunctionCode = Convert.ToInt16(str.Remove(str.IndexOf("\r\n")));
                    Console.WriteLine("Function Code: " + FunctionCode.ToString("X"));
                    MessageBox.Show("Function Code: " + FunctionCode.ToString("X"));

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);
                    ExceptionCode = str.Remove(str.IndexOf("-"));
                    switch (ExceptionCode.Trim())
                    {
                        case "1":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            break;
                        case "2":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            break;
                        case "3":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            break;
                        case "4":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            break;
                    }
                }
            }
        }

        private void reset_Click(object sender, EventArgs e)
        {
            try
            {
               master.WriteSingleRegister(1, 9, 0x55AA);
               reset.Enabled = true;
            }
            catch (Exception exception)
            {
                //Connection exception
                //No response from server.
                //The server maybe close the com port, or response timeout.
                if (exception.Source.Equals("System"))
                {
                    Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                    MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                    StartTest.Enabled = false;
                    reset.Enabled = false;
                    this.UpdateBtn.Enabled = false;
                    this.choosebth.Enabled = false;
                    return;
                }
                //The server return error code.
                //You can get the function code and exception code.
                if (exception.Source.Equals("nModbusPC"))
                {
                    string str = exception.Message;
                    int FunctionCode;
                    string ExceptionCode;

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);
                    FunctionCode = Convert.ToInt16(str.Remove(str.IndexOf("\r\n")));
                    Console.WriteLine("Function Code: " + FunctionCode.ToString("X"));
                    MessageBox.Show("Function Code: " + FunctionCode.ToString("X"));

                    str = str.Remove(0, str.IndexOf("\r\n") + 17);
                    ExceptionCode = str.Remove(str.IndexOf("-"));
                    switch (ExceptionCode.Trim())
                    {
                        case "1":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                            break;
                        case "2":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                            break;
                        case "3":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                            break;
                        case "4":
                            Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                            break;
                    }
                }
            }


        }
        private void bgWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            this.upgrade();
        }
        private void bgWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            UpdateBar.Value = e.ProgressPercentage;
        }
        private void choosebth_Click(object sender, EventArgs e)
        {
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.Multiselect = false;//该值确定是否可以选择多个文件
            dialog.Title = "请选择文件夹";
            dialog.Filter = "镜像文件(*.bin)|*.bin";
            if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                string file = dialog.FileName;
                this.ImagefilePath = file;
                this.imagesname.Text = file.Substring(file.LastIndexOf("\\") + 1); ;
            }
        }
        private void UpdateBtn_Click(object sender, EventArgs e)
        {
            bStart = true;
            int iMaxBufSize = 1024 * 100;
            this.UpdateBtn.Enabled = false;
            this.choosebth.Enabled = false;
            int iFileSize = 0;
            Byte[] FileBuf = new Byte[iMaxBufSize];

            if (ImagefilePath != null)
            {
                FileStream fsl = File.Open(ImagefilePath, FileMode.Open, FileAccess.ReadWrite);
                if (File.Exists(ImagefilePath) == true)
                {
                    BinaryReader br = new BinaryReader(fsl);

                    FileBuf = br.ReadBytes(iMaxBufSize);

                    iFileSize = FileBuf.Length;

                    br.Close();
                    fsl.Close();
                }
            }
            else
            {
                MessageBox.Show("程序不存在");
                this.UpdateBtn.Enabled = true;
                this.choosebth.Enabled = true;
            }

            UpdateBar.Maximum = iFileSize;
            UpdateBar.Minimum = 0;


            if (bStart == true)
            {
                this.bgWorker.WorkerReportsProgress = true;
                this.bgWorker.RunWorkerAsync();
                bStart = false;
            }
            else
            {
                this.bgWorker.WorkerReportsProgress = false;
            }
        }
        private void UpgradeProc()
        {
            while (true)
            {
                if (exitEvent.WaitOne(waitTime))
                {
                    break;
                }
                invokeDelegate setText = () =>
                {
                    if (this.UpdateBar.Value == this.UpdateBar.Maximum)
                    {
                        MessageBox.Show("更新成功！");
                        this.UpdateBar.Value = this.UpdateBar.Minimum;
                        this.UpdateBtn.Enabled = false;
                        this.choosebth.Enabled = true;
                    }
                };
                Invoke(setText);
            }
        }
        private void upgrade()
        {
            //CheckForIllegalCrossThreadCalls = false;
            ////设置CheckForIllegalCrossThreadCalls = false 是能去掉这个检查的,但是这不是标准的作法.
            ////标准的作法是在访问界面控件时,访问下窗体的 InvokeRequired 属性,如果为false就可以直接访问,
            ////否则就是跨线程访问; 此时,创建一个delegate,并通过Invoke() 来调用它.
            ////参考https://www.cnblogs.com/janghe/p/7874889.html
  
            const int Singlepackage =100;
            const int PerBlockPacket = Singlepackage;
            const int MB_ENOERR = 0;
            int iFileSizeNow = 0;
            int iFileSize = 0;
            int iMaxBufSize = 1024 * 100;
            ushort[] istage = new ushort[2];
            //ushort[] data = new ushort[2];

            Byte[] FileBuf = new Byte[iMaxBufSize];
            string strLog = "";
            UInt16 u16DevStatus = 0;
            ushort[] usTrans = new ushort[Singlepackage];

            UInt16[] u16unlock = { 0x5749, 0x5449, 0x554D };
            int ret;
            int errCnt = 0;

            FileStream fsl = File.Open(ImagefilePath, FileMode.Open, FileAccess.ReadWrite);
            if (File.Exists(ImagefilePath) == true)
            {
                BinaryReader br = new BinaryReader(fsl);

                FileBuf = br.ReadBytes(iMaxBufSize);

                iFileSize = FileBuf.Length;

                br.Close();
                fsl.Close();
            }
            else
            {
                MessageBox.Show("程序不存在");
            }



            istage[0] = 0;

            master.WriteSingleRegister(1, 20600, 0);

            ushort[] usFileBuf = new ushort[iMaxBufSize];
            Buffer.BlockCopy(FileBuf, 0, usFileBuf, 0, FileBuf.Length);
            bgWorker.ReportProgress(0);
            while (true)
            {
                try
                {
                    ushort[] data = master.ReadHoldingRegisters(1, 20600, 1);
                    u16DevStatus = data[0];
                    int sent = istage[0] * (Singlepackage * 2);

                    strLog = sent.ToString() + " has been sent.";
                    // UpdateBar.Value = sent;
                    bgWorker.ReportProgress(sent);

                    if (u16DevStatus == istage[0])
                    {

                    }
                    else
                    {
                        continue;
                    }

                    if (iFileSizeNow + PerBlockPacket * 2 <= iFileSize)
                    {
                        Buffer.BlockCopy(usFileBuf, iFileSizeNow, usTrans, 0, Singlepackage*2);
                        try
                        {
                            master.WriteMultipleRegisters(1, 20480, usTrans);
                            istage[0]++;
                            iFileSizeNow += PerBlockPacket * 2;
                            //Thread.Sleep(10);
                        }
                        catch (Exception exception)
                        {
                            //Connection exception
                            //No response from server.
                            //The server maybe close the com port, or response timeout.
                            if (exception.Source.Equals("System"))
                            {
                                if (errCnt > 5)
                                    return;
                                else
                                    errCnt++;

                                Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                                MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                                this.UpdateBar.Value = this.UpdateBar.Minimum;
                                this.StartTest.Enabled = false;
                                this.reset.Enabled = false;
                                this.UpdateBtn.Enabled = false;
                                this.choosebth.Enabled = false;
                                return;
                            }
                        }

                    }
                    else
                    {
                        //最后一帧
                        Array.Clear(usTrans, 0, usTrans.Length);
                        Buffer.BlockCopy(usFileBuf, iFileSizeNow, usTrans, 0, iFileSize - iFileSizeNow);
                        try
                        {
                            master.WriteMultipleRegisters(1, 20480, usTrans);
                            istage[0]++;
                            iFileSizeNow += PerBlockPacket * 2;
                            istage[0] = 0x5754;
                            bgWorker.ReportProgress(iFileSize);
                            while (true)
                            {
                                //ret = mb_dev.WriteMultipleRegs(wtdslaveid, 2173, 3, "", 502, ref u16unlock);
                                try
                                {
                                    master.WriteMultipleRegisters(1, 20601, u16unlock);
                                    break;
                                }
                                catch (Exception exception)
                                {
                                    //Connection exception
                                    //No response from server.
                                    //The server maybe close the com port, or response timeout.
                                    if (exception.Source.Equals("System"))
                                    {
                                        Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                                        MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                                        this.UpdateBar.Value = this.UpdateBar.Minimum;
                                        this.StartTest.Enabled = false;
                                        this.reset.Enabled = false;
                                        this.UpdateBtn.Enabled = false;
                                        this.choosebth.Enabled = false;
                                        return;
                                    }
                                }

                            }
                            return;
                        }
                        catch (Exception exception)
                        {
                            //Connection exception
                            //No response from server.
                            //The server maybe close the com port, or response timeout.
                            if (exception.Source.Equals("System"))
                            {
                                if (errCnt > 5)
                                    return;
                                else
                                    errCnt++;

                                Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                                MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                                this.UpdateBar.Value = this.UpdateBar.Minimum;
                                this.StartTest.Enabled = false;
                                this.reset.Enabled = false;
                                this.UpdateBtn.Enabled = false;
                                this.choosebth.Enabled = false;
                                return;
                            }
                        }
                    }

                }
                catch (Exception exception)
                {
                    //Connection exception
                    //No response from server.
                    //The server maybe close the com port, or response timeout.
                    if (exception.Source.Equals("System"))
                    {
                        if (errCnt > 5)
                            return;
                        else
                            errCnt++;

                        Console.WriteLine(DateTime.Now.ToString() + " " + exception.Message);
                        MessageBox.Show(DateTime.Now.ToString() + " " + exception.Message);
                        this.UpdateBar.Value = this.UpdateBar.Minimum;
                        this.StartTest.Enabled = false;
                        this.reset.Enabled = false;
                        this.UpdateBtn.Enabled = false;
                        this.choosebth.Enabled = false;
                        return;
                    }
                    //The server return error code.
                    //You can get the function code and exception code.
                    if (exception.Source.Equals("nModbusPC"))
                    {
                        string str = exception.Message;
                        int FunctionCode;
                        string ExceptionCode;

                        str = str.Remove(0, str.IndexOf("\r\n") + 17);
                        FunctionCode = Convert.ToInt16(str.Remove(str.IndexOf("\r\n")));
                        Console.WriteLine("Function Code: " + FunctionCode.ToString("X"));
                        MessageBox.Show("Function Code: " + FunctionCode.ToString("X"));

                        str = str.Remove(0, str.IndexOf("\r\n") + 17);
                        ExceptionCode = str.Remove(str.IndexOf("-"));
                        switch (ExceptionCode.Trim())
                        {
                            case "1":
                                Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                                MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal function!");
                                break;
                            case "2":
                                Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                                MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data address!");
                                break;
                            case "3":
                                Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                                MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Illegal data value!");
                                break;
                            case "4":
                                Console.WriteLine("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                                MessageBox.Show("Exception Code: " + ExceptionCode.Trim() + "----> Slave device failure!");
                                break;

                        }
                        this.UpdateBar.Value = this.UpdateBar.Minimum;
                        this.UpdateBtn.Enabled = false;
                        this.choosebth.Enabled = true;
                        return;
                    }
                }
            }
        }

        /****************转换函数*****************/
        public static bool[] GetBoolArr(byte b)
        {
            bool[] Bool = new bool[8];
            int i = 0;//读取出bit和byte顺序是颠倒的,利用i来作为数组下标
            for (int num = 7; num >= 0; num--)
            {
                Bool[i] = (1 << num & b) > 0;
                i++;
            }
            return Bool;
        }
        public static byte ConvertBCDToInt(byte b)
        {
            //高四位  
            byte b1 = (byte)((b >> 4) & 0xF);
            //低四位  
            byte b2 = (byte)(b & 0xF);

            return (byte)(b1 * 10 + b2);
        }

        public static byte[] HexStringToByteArray(string s)
        {
            s = s.Replace(" ", "");
            byte[] buffer = new byte[s.Length / 2];
            for (int i = 0; i < s.Length; i += 2)
                buffer[i / 2] = (byte)Convert.ToByte(s.Substring(i, 2), 16);
            return buffer;
        }
        //字符串转16进制字节组：
        public static byte[] strToHexByte(string hexString)
        {
            hexString = hexString.Replace(" ", "");   //去除空格
            if ((hexString.Length % 2) != 0)     //判断hexstring的长度是否为偶数
            {
                hexString += "";
            }
            //声明一个长度为hexstring长度一半的字节组returnBytes
            byte[] returnBytes = new byte[hexString.Length / 2];

            for (int i = 0; i < returnBytes.Length; i++)
            {
                //将hexstring的两个字符转换成16进制的字节组
                returnBytes[i] = Convert.ToByte(hexString.Substring(i * 2, 2), 16);
            }
            return returnBytes;
        }
        private string StringToHexString(string s, Encoding encode)
        {
            byte[] b = encode.GetBytes(s);//按照指定编码将string编程字节数组
            string result = string.Empty;
            for (int i = 0; i < b.Length; i++)//逐字节变为16进制字符，以%隔开
            {
                result += "%" + Convert.ToString(b[i], 16);
            }
            return result;
        }
        //字符串转16进制字节数组
        /// <summary>
        /// 字符串转16进制字节数组
        /// </summary>
        /// <param name="hexString"></param>
        /// <returns></returns>
        private static byte[] strToToHexByte(string hexString)
        {
            hexString = hexString.Replace(" ", "");
            if ((hexString.Length % 2) != 0)
                hexString = hexString.Insert(hexString.Length - 1, 0.ToString());
            byte[] returnBytes = new byte[hexString.Length / 2];
            for (int i = 0; i < returnBytes.Length; i++)
                returnBytes[i] = Convert.ToByte(hexString.Substring(i * 2, 2), 16);
            return returnBytes;
        }
        private void button2_Click_1(object sender, EventArgs e)
        {
            string sntext;
            string Names;
            int ki, kj;
            ushort[] vOut2= new ushort[4];
          
            UInt16[] SNIN16 = new UInt16[2];

            sntext = DTUSN.Text.ToString();
            int.TryParse(sntext, out ki);
            string s = ki.ToString("X");

            kj = int.Parse(s, System.Globalization.NumberStyles.AllowHexSpecifier);
            var m = (kj >> 16) & 0Xffff;
            var l = kj & 0xffff;
            SNIN16[0] = (ushort)m;
            SNIN16[1] = (ushort)l;
            master.WriteMultipleRegisters(1, 0, SNIN16);

            ushort[] SNC = master.ReadHoldingRegisters(1, 0, 3);
            SN = (SNC[0] << 16) + (Int32)SNC[1];
            SNcheck.Text = SN.ToString();



            Names = Nametype.Text.ToString().PadRight(8,' ');
                                             //int x = 0;
                                             //List<short> result = new List<short>();
                                             //while (NamesToHexByte.Length - x > 2)
                                             //{
                                             //    result.Add(BitConverter.ToInt16(NamesToHexByte, x));
                                             //    x += 2;
                                             //}
                                             //short[] arr = result.ToArray();//这就是你要的

            byte[] vOut = System.Text.Encoding.UTF8.GetBytes(Names);

            
            vOut2[0]= BitConverter.ToUInt16(vOut, 0 /* Which byte position to convert */);
            vOut2[1] = BitConverter.ToUInt16(vOut, 2 /* Which byte position to convert */);
            vOut2[2] = BitConverter.ToUInt16(vOut, 4 /* Which byte position to convert */);
            vOut2[3] = BitConverter.ToUInt16(vOut, 6 /* Which byte position to convert */);
            master.WriteMultipleRegisters(1, 3, vOut2);






            ushort[] register = master.ReadHoldingRegisters(1, 0, 3);
            SN = (register[0] << 16) + (Int32)register[1];
            DTUSN.Text = SN.ToString();

            ushort[] Namestr = master.ReadHoldingRegisters(1, 3, 6);
            for (int i = 0; i < 4; i++)
            {
                if (i < 2)
                    MNames[i] = (byte)((Namestr[0] >> (i * 8)) & 0x00ff);
                else
                    MNames[i] = (byte)(Namestr[1] >> ((i % 2) * 8) & 0x00ff);
            }
            for (int i = 4; i < 8; i++)
            {
                if (i < 6)
                    MNames[i] = (byte)((Namestr[2] >> (i * 8)) & 0x00ff);
                else
                    MNames[i] = (byte)(Namestr[3] >> ((i % 2) * 8) & 0x00ff);
            }
            for (int i = 8; i < 12; i++)
            {
                if (i < 10)
                    MNames[i] = (byte)((Namestr[4] >> (i * 8)) & 0x00ff);
                else
                    MNames[i] = (byte)(Namestr[5] >> ((i % 2) * 8) & 0x00ff);
            }

            Nametype.Text = System.Text.Encoding.ASCII.GetString(MNames);
        }

        private void DTUSN_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar != 8 && !Char.IsDigit(e.KeyChar))//如果不是输入数字就不让输入
            {
                e.Handled = true;
            }
        }
    }
}
